#include <stdio.h>
#include <stdlib.h>

#include <png.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spPluginP.h>

#include <sp/spDIBitmapP.h>

#define SP_NUM_PNG_TYPE 1
#define SP_PNG_TYPE_DEFAULT 0

typedef struct _spPngPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;

    FILE *fp;
    int sig_offset;
    spBool write_flag;
    png_structp png_ptr;
    png_infop info_ptr;
    png_infop end_info_ptr;

    double screen_gamma;
    double gamma;

    spDIBitmapInfo file_dibinfo;
    unsigned long file_strict_mask;
} *spPngPluginInstance;

#define SP_PNG_SIGNATURE_CMP_SIZE 8

static spBool spIsSupportedByPluginPng(const char *filename)
{
    spBool supported = SP_FALSE;

    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (strcaseeq(filename, ".png")) {
	    supported = SP_TRUE;
	}
    } else {
	FILE *fp;
	unsigned char header[SP_PNG_SIGNATURE_CMP_SIZE];
	int size = SP_PNG_SIGNATURE_CMP_SIZE;

	if ((fp = spOpenFile(filename, "rb")) != NULL) {
	    if (fread(header, 1, size, fp) == size && png_sig_cmp(header, 0, size) == 0) {
		supported = SP_TRUE;
	    }
	    spCloseFile(fp);
	}
    }
    
    return supported;
}

static spBool spInitPluginPng(const char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginPng(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstancePng(const char *lang)
{
    spPngPluginInstance pinstance;
    
    pinstance = xalloc(1, struct _spPngPluginInstance);
    memset(pinstance, 0, sizeof(struct _spPngPluginInstance));
    pinstance->type_index = SP_PNG_TYPE_DEFAULT;
    pinstance->screen_gamma = 2.2;
    pinstance->gamma = 1.0 / 2.2;
    
    return (void *)pinstance;
}

static spBool spFreePluginInstancePng(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypePng(void *instance, int index)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    if (index == 0) {
	pinstance->type_index = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static int spGetPluginFileTypePng(void *instance)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    return pinstance->type_index;
}

static const char *spGetPluginBestSuffixPng(void *instance)
{
    return ".png";
}

static unsigned long spGetPluginDIBCapsPng(void *instance)
{
    return SP_DI_BITMAP_CAPS_DIRECT_ACCESS | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3 | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4
	| SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_UPSIDE_UP | SP_DI_BITMAP_CAPS_UPSIDE_DOWN 
	| SP_DI_BITMAP_CAPS_NONINVERSE_FORMAT | SP_DI_BITMAP_CAPS_INVERSE_FORMAT
	| SP_DI_BITMAP_CAPS_ALPHA | SP_DI_BITMAP_CAPS_FIRST_ALPHA | SP_DI_BITMAP_CAPS_LAST_ALPHA
	| SP_DI_BITMAP_CAPS_GRAYSCALE;
}

static spBool spClosePluginPng(void *instance)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    if (pinstance->png_ptr != NULL) {
	if (pinstance->write_flag) {
	    png_destroy_write_struct(&pinstance->png_ptr, &pinstance->info_ptr);
	} else {
	    png_destroy_read_struct(&pinstance->png_ptr, &pinstance->info_ptr, &pinstance->end_info_ptr);
	}
    }

    pinstance->png_ptr = NULL;
    pinstance->info_ptr = NULL;
    pinstance->end_info_ptr = NULL;
    
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
    }
    
    return SP_TRUE;
}

static spPluginError spOpenPluginPng(void *instance, const char *filename, const char *mode)
{
    const char *mode2;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    if (mode[0] == 'w') {
	mode2 = "wb";
    } else {
	mode2 = "rb";
    }
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, mode2)) == NULL) {
	spDebug(10, "spOpenPluginPng", "can't open file: %s\n", filename);
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(80, "spOpenPluginPng", "filename = %s, mode = %s\n", filename, mode);

    pinstance->png_ptr = NULL;
    pinstance->info_ptr = NULL;
    pinstance->end_info_ptr = NULL;

    if (mode[0] == 'w') {
	pinstance->write_flag = SP_TRUE;
	pinstance->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	pinstance->sig_offset = 0;
    } else {
	unsigned char header[SP_PNG_SIGNATURE_CMP_SIZE];
	
	pinstance->sig_offset = SP_PNG_SIGNATURE_CMP_SIZE;

	if (fread(header, 1, pinstance->sig_offset, pinstance->fp) != pinstance->sig_offset
	    || png_sig_cmp(header, 0, pinstance->sig_offset) != 0) {
	    spCloseFile(pinstance->fp);
	    pinstance->fp = NULL;
	    spDebug(10, "spOpenPluginPng", "can't read png signature\n");
	    return SP_PLUGIN_ERROR_OPEN;
	}
	
	pinstance->write_flag = SP_FALSE;
	pinstance->png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    }

    if (pinstance->png_ptr != NULL) {
	pinstance->info_ptr = png_create_info_struct(pinstance->png_ptr);
    }

    if (pinstance->info_ptr == NULL) {
	spClosePluginPng(pinstance);
	spDebug(10, "spOpenPluginPng", "libpng initialize error\n");
	return SP_PLUGIN_ERROR_OPEN;
    }
    
    if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	spDebug(10, "spOpenPluginPng", "error\n");
	spClosePluginPng(pinstance);
	return SP_PLUGIN_ERROR_OPEN;
    }
    
    png_init_io(pinstance->png_ptr, pinstance->fp);
    
    if (pinstance->sig_offset > 0) {
	png_set_sig_bytes(pinstance->png_ptr, pinstance->sig_offset);
	png_read_info(pinstance->png_ptr, pinstance->info_ptr);
	
	png_get_gAMA(pinstance->png_ptr, pinstance->info_ptr, &pinstance->gamma);
	spDebug(80, "spOpenPluginPng", "gamma = %f\n", pinstance->gamma);
    }

    spDebug(80, "spOpenPluginPng", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spSetPluginOtherInfoPng(void *instance, const char *id, char *data)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    if (streq(id, SP_PLUGIN_INFO_DIBITMAP_GAMMA)) {
	double *gamma = (double *)data;
	pinstance->gamma = *gamma;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoPng(void *instance, const char *id, char *data)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    if (streq(id, SP_PLUGIN_INFO_DIBITMAP_GAMMA)) {
	double *gamma = (double *)data;
	*gamma = pinstance->gamma;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spSetPluginFileDIBInfoPng(void *instance, spDIBitmapInfo *dibinfo, unsigned long *strict_mask)
{
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    if (pinstance->write_flag == SP_FALSE) return SP_FALSE;
    
    spDebug(80, "spSetPluginFileDIBInfoPng", "bit_size = %ld, pixel_format = %lx\n",
	    dibinfo->bit_size, dibinfo->pixel_format);
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
	) {
	spDebug(10, "spSetPluginFileDIBInfoPng", "wrong format\n");
	return SP_FALSE;
    }

    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK) {
	dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_YUV_MASK | SP_DI_PIXEL_FORMAT_OPTIONAL_PARAMETERS_MASK);
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
    }
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }
    
    *strict_mask |= SP_DI_BITMAP_STRICT_YUV_STATE | SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE
	| SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE | SP_DI_BITMAP_STRICT_FLOAT_STATE;
    
    memcpy(&pinstance->file_dibinfo, dibinfo, sizeof(spDIBitmapInfo));
    pinstance->file_strict_mask = *strict_mask;

    spDebug(100, "spSetPluginFileDIBInfoPng", "done\n");
    
    return SP_TRUE;
}

static int spGetColorTypePng(spDIPixelFormat pixel_format)
{
    int color_type = 0;

    if (pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
	color_type = PNG_COLOR_TYPE_GRAY; /* 0 */
    } else {
	color_type = PNG_COLOR_MASK_COLOR;
    }
    
    if (pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	color_type |= PNG_COLOR_MASK_ALPHA;
    }

    return color_type;
}

static spBool spSetPluginDIBInfoPng(void *instance, spDIBitmapInfoInternal *dibinfo)
{
    int bit_depth;
    int color_type;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    if (pinstance->write_flag == SP_FALSE) {
        if ((dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK)
	    || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK)
	    || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)
            ) {
            spDebug(10, "spSetPluginDIBInfoPng", "unsupported pixel format: %lx\n", dibinfo->pixel_format);
            return SP_FALSE;
        }

	if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
            if (dibinfo->pixel_stride != 1) {
                return SP_FALSE;
            }
	} else if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)  {
            if (dibinfo->pixel_stride != 4) {
                return SP_FALSE;
            }
	} else {
            if (dibinfo->pixel_stride != 3) {
                return SP_FALSE;
            }
        }
        
        return SP_TRUE;
    }

    if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	spDebug(10, "spSetPluginDIBInfoPng", "png library error\n");
	return SP_FALSE;
    }
    
    if (pinstance->file_dibinfo.pixel_format) {
	bit_depth = pinstance->file_dibinfo.bit_size;
	color_type = spGetColorTypePng(pinstance->file_dibinfo.pixel_format);
    } else {
	bit_depth = dibinfo->bit_size;
	color_type = spGetColorTypePng(dibinfo->pixel_format);
    }
    spDebug(80, "spSetPluginDIBInfoPng", "bit_depth = %d, color_type = %d\n", bit_depth, color_type);

    png_set_IHDR(pinstance->png_ptr, pinstance->info_ptr, dibinfo->width, dibinfo->height,
		 bit_depth, color_type, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    if (bit_depth == 16) {
	if (dibinfo->bit_size == 8) {
	    png_set_strip_16(pinstance->png_ptr);
	}
    } else if (bit_depth < 8) {
        png_set_packing(pinstance->png_ptr);
    }

    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	png_set_bgr(pinstance->png_ptr);
    }
    
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	if (!(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK)) {
	    png_set_swap_alpha(pinstance->png_ptr);
	}
    } else {
	if (dibinfo->pixel_stride == 4
	    || ((dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)
		&& dibinfo->pixel_stride == 2))  {
	    png_set_filler(pinstance->png_ptr, 0,
			   (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) ? PNG_FILLER_BEFORE : PNG_FILLER_AFTER);
	}
    }

#if BYTE_ORDER == LITTLE_ENDIAN
    if (dibinfo->bit_size > 8) {
	png_set_swap(pinstance->png_ptr);
    }
#endif

    png_set_gamma(pinstance->png_ptr, pinstance->screen_gamma, pinstance->gamma);

    return SP_TRUE;
}
    
static spBool spGetPluginFileDIBInfoPng(void *instance, spDIBitmapInfo *dibinfo)
{
    png_uint_32 width;
    png_uint_32 height;
    int bit_depth;
    int color_type;
    int interlace_method;
    int compression_method;
    int filter_method;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    if (pinstance->write_flag) return SP_FALSE;

    spDebug(80, "spGetPluginFileDIBInfoPng", "in\n");
    
    if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	spDebug(10, "spGetPluginFileDIBInfoPng", "error\n");
	return SP_FALSE;
    }
    
    png_get_IHDR(pinstance->png_ptr, pinstance->info_ptr, &width, &height,
		 &bit_depth, &color_type, &interlace_method,
		 &compression_method, &filter_method);
    spDebug(80, "spGetPluginFileDIBInfoPng", "width = %d, height = %d, bit_depth = %d, color_type = %d\n",
	    width, height, bit_depth, color_type);

    dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_YUV_MASK | SP_DI_PIXEL_FORMAT_OPTIONAL_PARAMETERS_MASK
			       | SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK | SP_DI_PIXEL_FORMAT_FLOAT_MASK);
    
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }
    
    if (color_type == PNG_COLOR_TYPE_PALETTE) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
    } else if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_RGB_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
    } else {
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
    }
    
    if (color_type & PNG_COLOR_MASK_ALPHA) {
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK | SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    } else {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }
    
#if BYTE_ORDER == LITTLE_ENDIAN
    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_SWAP_ENDIAN_PIXEL_MASK;
#endif
    
    dibinfo->width = width;
    dibinfo->height = height;

    dibinfo->bit_size = bit_depth;
    dibinfo->upside_down = SP_FALSE;
    spDebug(80, "spGetPluginFileDIBInfoPng", "pixel_format = %lx, bit_size = %ld\n", dibinfo->pixel_format, dibinfo->bit_size);
    
    spDebug(80, "spGetPluginFileDIBInfoPng", "done\n");
    
    return SP_TRUE;
}

static spBool spGetPluginDIBInfoPng(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned long *strict_mask)
{
    png_uint_32 width;
    png_uint_32 height;
    int bit_depth;
    int color_type;
    int interlace_method;
    int compression_method;
    int filter_method;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    spDebug(80, "spGetPluginDIBInfoPng", "in\n");
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
	) {
	spDebug(10, "spGetPluginDIBInfoPng", "unsupported strict mask\n");
	return SP_FALSE;
    }
    
    dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_YUV_MASK | SP_DI_PIXEL_FORMAT_OPTIONAL_PARAMETERS_MASK
			       | SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK | SP_DI_PIXEL_FORMAT_FLOAT_MASK);
    
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }

    if (pinstance->write_flag) {
	if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 1) {
		spDebug(10, "spGetPluginDIBInfoPng", "not supported pixel_stride for grayscale\n");
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 1;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_GRAY;
	} else if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)  {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 4) {
		spDebug(10, "spGetPluginDIBInfoPng", "not supported pixel_stride for rgba\n");
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 4;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_RGBA;
	} else {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 3) {
		spDebug(10, "spGetPluginDIBInfoPng", "not supported pixel_stride for rgb\n");
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 3;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
	}
	
	if ((*strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE)
	    && dibinfo->line_stride != dibinfo->width * dibinfo->pixel_stride) {
	    spDebug(10, "spGetPluginDIBInfoPng", "not supported line_stride\n");
	    return SP_FALSE;
	} else {
	    dibinfo->line_stride = dibinfo->width * dibinfo->pixel_stride;
	}
	dibinfo->bit_size = 8;
	
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_FALSE;
	}
    } else {
	if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	    spDebug(10, "spGetPluginDIBInfoPng", "error\n");
	    return SP_FALSE;
	}
    
	png_get_IHDR(pinstance->png_ptr, pinstance->info_ptr, &width, &height,
		     &bit_depth, &color_type, &interlace_method,
		     &compression_method, &filter_method);
	spDebug(80, "spGetPluginDIBInfoPng", "width = %d, height = %d, bit_depth = %d, color_type = %d\n",
		width, height, bit_depth, color_type);

	if (color_type == PNG_COLOR_TYPE_PALETTE) {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
		&& (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)) {
		return SP_FALSE;
	    }
	    png_set_palette_to_rgb(pinstance->png_ptr);
	    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
	    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
	} else if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
		&& (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)) {
		dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_RGB_MASK;
		dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
	    } else {
		png_set_gray_to_rgb(pinstance->png_ptr);
		dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
		dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
	    }
	} else {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
		&& (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)) {
		return SP_FALSE;
	    }
	    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
	}

#if BYTE_ORDER == LITTLE_ENDIAN
	if (dibinfo->bit_size > 8) {
	    png_set_swap(pinstance->png_ptr);
	}
#endif
    
	if (bit_depth == 16) {
	    if (!(*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)) {
		png_set_strip_16(pinstance->png_ptr);
		bit_depth = 8;
	    }
	} else if (bit_depth < 8) {
	    png_set_packing(pinstance->png_ptr);
	    bit_depth = 8;
	}

	if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	    png_set_bgr(pinstance->png_ptr);
	} else {
	    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_INVERSE_MASK;
	}

	if (color_type & PNG_COLOR_MASK_ALPHA) {
	    spDebug(80, "spGetPluginDIBInfoPng", "alpha channel included\n");
	
	    if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		 && dibinfo->pixel_stride == 3)
		|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)))  {
		spDebug(80, "spGetPluginDIBInfoPng", "call png_set_strip_alpha\n");
		png_set_strip_alpha(pinstance->png_ptr);
		dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_MASK;
	    
		if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		    && dibinfo->pixel_stride == 4) {
		    spDebug(80, "spGetPluginDIBInfoPng", "call png_set_filler\n");
		    png_set_filler(pinstance->png_ptr, 0,
				   (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) ? PNG_FILLER_BEFORE : PNG_FILLER_AFTER);
		} else {
		    dibinfo->pixel_stride = 3;
		}
	    } else {
		dibinfo->pixel_stride = 4;
		if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
		    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK)) {
		    spDebug(80, "spGetPluginDIBInfoPng", "call png_set_swap_alpha\n");
		    png_set_swap_alpha(pinstance->png_ptr);
		    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK | SP_DI_PIXEL_FORMAT_ALPHA_MASK;
		} else {
		    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK;
		    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK | SP_DI_PIXEL_FORMAT_ALPHA_MASK;
		}
	    }
	} else {
	    spDebug(80, "spGetPluginDIBInfoPng", "no alpha channel\n");
	
	    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_MASK;
	
	    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
		dibinfo->pixel_stride = 1;
	    } else if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		       && dibinfo->pixel_stride == 4) {
		spDebug(80, "spGetPluginDIBInfoPng", "call png_set_filler\n");
		png_set_filler(pinstance->png_ptr, 0,
			       (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) ? PNG_FILLER_BEFORE : PNG_FILLER_AFTER);
	    } else {
		dibinfo->pixel_stride = 3;
	    }
	
	}

	dibinfo->width = width;
	dibinfo->height = height;

	dibinfo->bit_size = bit_depth;
	dibinfo->line_stride = dibinfo->width * dibinfo->pixel_stride;
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_FALSE;
	}
	spDebug(80, "spGetPluginDIBInfoPng", "pixel_format = %lx, bit_size = %ld, pixel_stride = %ld, line_stride = %ld\n",
		dibinfo->pixel_format, dibinfo->bit_size, dibinfo->pixel_stride, dibinfo->line_stride);

	*strict_mask |= SP_DI_BITMAP_STRICT_YUV_STATE | SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE
	    | SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE | SP_DI_BITMAP_STRICT_FLOAT_STATE;
    }
    
    spDebug(80, "spGetPluginDIBInfoPng", "done\n");
    
    return SP_TRUE;
}

static spBool spReadPluginDIBDataPng(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    int p, i;
    int num_pass;
    long stride;
    png_bytep rowp;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;
    
    spDebug(80, "spReadPluginDIBDataPng", "in\n");
    
    if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	spDebug(10, "spReadPluginDIBDataPng", "error\n");
	return SP_FALSE;
    }

    png_read_update_info(pinstance->png_ptr, pinstance->info_ptr);

    num_pass = png_set_interlace_handling(pinstance->png_ptr);
    spDebug(80, "spReadPluginDIBDataPng", "num_pass = %d\n", num_pass);

    for (p = 0; p < num_pass; p++) {
	if (dibinfo->upside_down) {
	    rowp = (png_bytep)(data + dibinfo->line_stride * (dibinfo->height - 1));
	    stride = -dibinfo->line_stride;
	} else {
	    rowp = (png_bytep)data;
	    stride = dibinfo->line_stride;
	}
	
	for (i = 0; i < dibinfo->height; i++) {
	    png_read_row(pinstance->png_ptr, rowp, NULL);
	    rowp += stride;
	}
    }
    
    png_read_end(pinstance->png_ptr, pinstance->end_info_ptr);
    
    spDebug(80, "spReadPluginDIBDataPng", "done\n");
    
    return SP_TRUE;
}

static spBool spWritePluginDIBDataPng(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    int p, i;
    int num_pass;
    long stride;
    png_bytep rowp;
    spPngPluginInstance pinstance = (spPngPluginInstance)instance;

    spDebug(80, "spWritePluginDIBDataPng", "in\n");
    
    if (setjmp(png_jmpbuf(pinstance->png_ptr))) {
	spDebug(10, "spWritePluginDIBDataPng", "error\n");
	return SP_FALSE;
    }

    png_write_info(pinstance->png_ptr, pinstance->info_ptr);

    num_pass = png_set_interlace_handling(pinstance->png_ptr);
    spDebug(80, "spWritePluginDIBDataPng", "num_pass = %d\n", num_pass);

    for (p = 0; p < num_pass; p++) {
	if (dibinfo->upside_down) {
	    rowp = (png_bytep)(data + dibinfo->line_stride * (dibinfo->height - 1));
	    stride = -dibinfo->line_stride;
	} else {
	    rowp = (png_bytep)data;
	    stride = dibinfo->line_stride;
	}
    
	for (i = 0; i < dibinfo->height; i++) {
	    png_write_row(pinstance->png_ptr, rowp);
	    rowp += stride;
	}
    }
    
    png_write_end(pinstance->png_ptr, pinstance->info_ptr);

    spDebug(80, "spWritePluginDIBDataPng", "done\n");
    
    return SP_TRUE;
}

static const char *sp_png_file_type_list[] = {
    "png",
    NULL,
};
static const char *sp_png_file_desc_list[] = {
    "PNG (libpng)",
    NULL,
};
static const char *sp_png_file_filter_list[] = {
    "*.png",
    NULL,
};
    
static spDIBitmapPluginRec sp_dibitmap_plugin_png = {
    NULL,
    NULL,

    SP_PLUGIN_DIBITMAP,
    "png",
    1,
    SP_PLUGIN_PRIORITY_HIGHER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginPng,
    spFreePluginPng,
    "PNG I/O Plugin",
    "PNG I/O Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitPluginInstancePng,
    spFreePluginInstancePng,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_png_file_type_list,
    sp_png_file_desc_list,
    sp_png_file_filter_list,
    spIsSupportedByPluginPng,
    spSetPluginFileTypePng,  
    spGetPluginFileTypePng,
#if 0
    SP_META_TITLE_MASK | SP_META_AUTHOR_MASK | SP_META_DESCRIPTION_MASK | SP_META_WARNING_MASK
	| SP_META_RELEASE_MASK | SP_META_COPYRIGHT_MASK | SP_META_DISCLAIMER_MASK | SP_META_SOURCE_MASK
	| SP_META_SOFTWARE_MASK | SP_META_SUBJECT_MASK | SP_META_COMMENT_MASK,
    spSetPluginMetaInfoPng,  
    spGetPluginMetaInfoPng,
#else
    0L,
    NULL,
    NULL,
#endif
    spGetPluginBestSuffixPng,
    
    spGetPluginDIBCapsPng,
	
    spOpenPluginPng,              
    spClosePluginPng,
	
    spSetPluginFileDIBInfoPng,
    spSetPluginDIBInfoPng,
    spGetPluginFileDIBInfoPng,
    spGetPluginDIBInfoPng,
    spSetPluginOtherInfoPng,   
    spGetPluginOtherInfoPng,
	
    spReadPluginDIBDataPng,          
    spWritePluginDIBDataPng,
	
    NULL,
    NULL,
};

#if defined(SP_DIBITMAP_PLUGIN_USE_PNG) && (defined(GTK) || defined(USE_MOTIF))
spBool spAddPngDIBitmapPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_dibitmap_plugin_png, "png");
}

spBool spRemovePngDIBitmapPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_dibitmap_plugin_png);
}
#else
#include <sp/spPluginMain.h>

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_dibitmap_plugin_png;
}
#endif
