/*
 *	spDrawingPlugin.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDrawingPluginP.h>

#define SP_NUM_DRAWING_PLUGIN_BUFFER 8

typedef struct _spDrawingPluginFileInfo {
    char *name;
    char *description;
    char *information;
    spDrawingDeviceType device_type;
} spDrawingPluginFileInfo;

static int sp_num_plugin_buffer = 0;
static int sp_num_plugin_file = 0;
static spDrawingPluginFileInfo *sp_plugin_files = NULL;

static int initDrawingPluginFileInfoAll(void)
{
    int i;
    char *plugin_name;
    spPlugin *plugin;
    
    if (sp_plugin_files == NULL) {
	sp_num_plugin_buffer = SP_NUM_DRAWING_PLUGIN_BUFFER;
	sp_plugin_files = xalloc(sp_num_plugin_buffer, spDrawingPluginFileInfo);
	
	for (i = 0;; i++) {
	    if ((plugin_name = xspSearchPluginFile(i)) == NULL) {
		break;
	    }

	    if ((plugin = spLoadPlugin(plugin_name)) != NULL) {
		if (spIsDrawingPlugin(plugin) == SP_TRUE) {
		    sp_plugin_files[sp_num_plugin_file].name = strclone(plugin_name);
		    sp_plugin_files[sp_num_plugin_file].description = strclone(spGetPluginDescription(plugin));
		    sp_plugin_files[sp_num_plugin_file].information = strclone(spGetPluginInformation(plugin));
		    sp_plugin_files[sp_num_plugin_file].device_type = spGetPluginDrawingDeviceType(plugin);
		    sp_num_plugin_file++;

		    if (sp_num_plugin_file >= sp_num_plugin_buffer) {
			sp_num_plugin_buffer += SP_NUM_DRAWING_PLUGIN_BUFFER;
			sp_plugin_files = xrealloc(sp_plugin_files, sp_num_plugin_buffer, spDrawingPluginFileInfo);
		    }
		}
		spFreePlugin(plugin);
	    }
	    xfree(plugin_name);
	}
    }

    return sp_num_plugin_file;
}

static void freeDrawingPluginFileInfoAll(void)
{
    int i;
    
    for (i = 0; i < sp_num_plugin_file; i++) {
	if (sp_plugin_files[i].name != NULL) {
	    xfree(sp_plugin_files[i].name);
	}
	if (sp_plugin_files[i].description != NULL) {
	    xfree(sp_plugin_files[i].description);
	}
	if (sp_plugin_files[i].information != NULL) {
	    xfree(sp_plugin_files[i].information);
	}
    }
    xfree(sp_plugin_files);
    sp_num_plugin_file = 0;
    sp_plugin_files = NULL;

    return;
}

const char *spSearchDrawingPluginFile(int index, spDrawingDeviceType device_type)
{
    int i;
    int count;
    int num_file;
    
    if (index < 0) {
	freeDrawingPluginFileInfoAll();
	return NULL;
    }

    num_file = initDrawingPluginFileInfoAll();

    count = 0;
    for (i = 0; i < num_file; i++) {
	if ((sp_plugin_files[i].device_type & device_type) || device_type == SP_DRAWING_DEVICE_UNKNOWN) {
	    if (index == count) {
		return sp_plugin_files[i].name;
	    }
	    count++;
	}
    }

    return NULL;
}

char *xspSearchDrawingPluginFile(int index, spDrawingDeviceType device_type)
{
    return strclone(spSearchDrawingPluginFile(index, device_type));
}

const char *spGetDrawingPluginDescription(const char *plugin_name)
{
    int i;
    int num_file;

    if (strnone(plugin_name)) return NULL;
    
    num_file = initDrawingPluginFileInfoAll();

    for (i = 0; i < num_file; i++) {
	if (streq(sp_plugin_files[i].name, plugin_name)) {
	    return sp_plugin_files[i].description;
	}
    }

    return NULL;
}

char *xspGetDrawingPluginDescription(const char *plugin_name)
{
    return strclone(spGetDrawingPluginDescription(plugin_name));
}

const char *spGetDrawingPluginInformation(const char *plugin_name)
{
    int i;
    int num_file;

    if (strnone(plugin_name)) return NULL;
    
    num_file = initDrawingPluginFileInfoAll();

    for (i = 0; i < num_file; i++) {
	if (streq(sp_plugin_files[i].name, plugin_name)) {
	    return sp_plugin_files[i].information;
	}
    }

    return NULL;
}

char *xspGetDrawingPluginInformation(const char *plugin_name)
{
    return strclone(spGetDrawingPluginInformation(plugin_name));
}

spBool spIsDrawingPlugin(spPlugin *plugin)
{
    if (plugin == NULL) return SP_FALSE;

    if (spEqPluginType(plugin, SP_PLUGIN_DRAWING) == SP_TRUE) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

spDrawingDeviceType spGetPluginDrawingDeviceType(spPlugin *plugin)
{
    if (plugin == NULL) return SP_DRAWING_DEVICE_UNKNOWN;

    if (spIsDrawingPlugin(plugin) == SP_TRUE) {
	return ((spDrawingPluginRec *)plugin->rec)->device_type;
    }
    
    return SP_DRAWING_DEVICE_UNKNOWN;
}

static spBool sp_drawing_plugin_load_status = SP_FALSE;

spBool spAddDefaultDrawingPlugin(void)
{
    if (sp_drawing_plugin_load_status == SP_FALSE) {
#if !(/*defined(IPHONE) ||*/ defined(ANDROID))
	if (spAddPrinterPlugin() == SP_TRUE) {
	    sp_drawing_plugin_load_status = SP_TRUE;
	}
#endif
	if (spAddPSPlugin() == SP_TRUE) {
	    sp_drawing_plugin_load_status = SP_TRUE;
	}
#if defined(_WIN32) && !(defined(GTK) || defined(USE_MOTIF))
	if (spAddEmfPlugin() == SP_TRUE) {
	    sp_drawing_plugin_load_status = SP_TRUE;
	}
#endif
    }
    
    return sp_drawing_plugin_load_status;
}

spBool spRemoveDefaultDrawingPlugin(void)
{
    spBool status = SP_FALSE;

    if (spRemovePSPlugin() == SP_TRUE) {
	status = SP_TRUE;
    }
#if !(/*defined(IPHONE) ||*/ defined(ANDROID))
    if (spRemovePrinterPlugin() == SP_TRUE) {
	status = SP_TRUE;
    }
#endif
#if defined(_WIN32) && !(defined(GTK) || defined(USE_MOTIF))
    if (spRemoveEmfPlugin() == SP_TRUE) {
	status = SP_TRUE;
    }
#endif

    sp_drawing_plugin_load_status = SP_FALSE;
    
    return status;
}

